//Line Follower https://www.instructables.com/id/Line-Follower-Basic-Using-Arduino-Nano/
//TCS 3200 https://howtomechatronics.com/projects/arduino-color-sorter-project/
//Intersection https://www.youtube.com/watch?v=Ux2L_KAkHMQ

#include <Servo.h>

Servo boxServo;

//Color sensor TCS3200
#define S0 2     
#define S1 3
#define S2 8
#define S3 9
#define sensorOut 10
int frequency = 0;
int color=0;

//Motor L298N
int mRight = 5; 
int dirRight = 4; 
int mLeft = 6; 
int dirLeft = 7;

//Line sensor
int sensorVal[6] = {0,0,0,0,0,0};
int error, last_error, MV,pid_l,pid_r,D,D1,D2,D3,I,I1,I2,I3,P,Pd, bitSensor;
int Max_MV;
unsigned char Kp = 40;
unsigned char Kd = 0;
unsigned char Ts = 1;
unsigned char maxPwm = 125;
unsigned char intersection = 0;

void setup(){
  Serial.begin(9600);
  //Motor L298N
  pinMode(mRight, OUTPUT); 
  pinMode(dirRight, OUTPUT); 
  pinMode(mLeft, OUTPUT); 
  pinMode(dirLeft, OUTPUT);  
  //Line sensor
  pinMode(A0, INPUT);
  pinMode(A1, INPUT);
  pinMode(A2, INPUT);
  pinMode(A3, INPUT);
  pinMode(A4, INPUT);
  pinMode(A5, INPUT);
  //Color sensor
  pinMode(S0, OUTPUT);
  pinMode(S1, OUTPUT);
  pinMode(S2, OUTPUT);
  pinMode(S3, OUTPUT);
  pinMode(sensorOut, INPUT);
  digitalWrite(S0,HIGH);
  digitalWrite(S1,LOW);  
  boxServo.attach(11);  
}

void(* resetFunc) (void) = 0;

void loop(){
  mission();
}

void readSens(){
  sensorVal[0] = digitalRead(A0);sensorVal[1] = digitalRead(A1); sensorVal[2] = digitalRead(A2);
  sensorVal[3] = digitalRead(A3); sensorVal[4] = digitalRead(A4); sensorVal[5] = digitalRead(A5);
}

void robotRun(){   
  readSens();
  bitSensor = ((sensorVal[0]*1)+(sensorVal[1]*2)+(sensorVal[2]*4)+(sensorVal[3]*8)+(sensorVal[4]*16)+(sensorVal[5]*32));
  switch(bitSensor){
    case 0b100000: error = -5; break;
    case 0b110000: error = -4; break;
    case 0b010000: error = -3; break;
    case 0b011000: error = -2; break;
    case 0b001000: error = -1; break;
    case 0b001100: error = 0; break;
    case 0b000100: error = 1; break;
    case 0b000110: error = 2; break;
    case 0b000010: error = 3; break;
    case 0b000011: error = 4; break;
    case 0b000001: error = 5; break;   
  }
  
  Max_MV = Kp*5; 
  P = Kp * error;
  D1 = Kd*8;
  D2 = D1 / Ts;
  D3 = error - last_error;
  D = D2 * D3;
 
  last_error = error;
  MV = P + D;
  
  if(MV>=-Max_MV && MV<=Max_MV){ 
    pid_l = maxPwm + MV;
    pid_r = maxPwm - MV;
    if (pid_l < 0) pid_l = 0;
    if (pid_l > 255) pid_l = 255;
    if (pid_r < 0) pid_r = 0;
    if (pid_r > 255) pid_r = 255;
    forward(pid_r,pid_l);
  }
  else if(MV<-Max_MV){
    turnLeft(200,100);
  }
  else if(MV>Max_MV){
    turnRight(100,200);
  }
  else{
    forward(pid_r,pid_l);
  }
}

int readColor(){
  //Red
  digitalWrite(S2, LOW);
  digitalWrite(S3, LOW);
  frequency = pulseIn(sensorOut, LOW);
  int R = frequency;
  Serial.print("R= ");
  Serial.print(frequency);
  Serial.print("  ");
  delay(50);
  //Green
  digitalWrite(S2, HIGH);
  digitalWrite(S3, HIGH);
  frequency = pulseIn(sensorOut, LOW);
  int G = frequency;
  Serial.print("G= ");
  Serial.print(frequency);
  Serial.print("  ");  
  delay(50);
  //Blue
  digitalWrite(S2, LOW);
  digitalWrite(S3, HIGH);
  frequency = pulseIn(sensorOut, LOW);
  int B = frequency;
  Serial.print("B= ");
  Serial.print(frequency);
  Serial.println("  ");  
  delay(50);

  if(R<64 & R>57 & G<70 & G>60){
    //Serial.println(" Red !\n");
    color = 1; //Red
  }
  if(R<110 & R>100 & G<57 & G>47){
    ///Serial.println(" Green !\n");
    color = 2; //Green
  }  
  if(G<56 & G>46 & B<156 & B>146){
    //Serial.println(" Blue !\n");
    color = 3; //Blue
  }
    return color;  
}

void mission(){
  robotRun();
  boxServo.write(10);
  if(sensorVal[5] == 1 && sensorVal[0] == 1){ 
  stopRun();   
  color=readColor();

  switch(color){
    case 1: //Turn Left 
      red();
      break;

    case 2: //Stright
      green();
      break;    
    
    case 3: //Turn Right
      blue();
      break;

    case 0:
      stopRun();
      break;  
    }
  }    
}

void red(){
  intersection++;
  if(intersection == 1){
    turnBack(85, 85); 
    }
    else if(intersection == 2){
      turnLeft(85, 85);
      }
    else if(intersection == 3){
      stopRun();
      boxServo.write(90);
      delay(1000);
      turnBack(85, 85);
      }
    else if(intersection == 4){
      turnRight(85, 85);      
      }
    else if(intersection == 5){ 
      resetFunc(); 
    }
}

void green(){
  intersection++;
  if(intersection == 1){   
    turnBack(85, 85);
    }
    else if(intersection == 2){
      stright(85, 85);
      }
    else if(intersection == 3){
      stopRun();
      boxServo.write(90);
      delay(1000);
      turnBack(85, 85);
      }
    else if(intersection == 4){ 
      stright(85, 85);     
      } 
    else if(intersection == 5){ 
      resetFunc(); 
    }  
}

void blue(){
  intersection++;
  if(intersection == 1){   
    turnBack(85, 85);
    }
    else if(intersection == 2){
      turnRight(85, 85); 
      }
    else if(intersection == 3){
      stopRun();
      boxServo.write(90);
      delay(1000);
      turnBack(85, 85);
      }
    else if(intersection == 4){
      turnLeft(85, 85);  
      }
    else if(intersection == 5){ 
      resetFunc(); 
    }  
}

void forward(int valLeft, int valRight){
  digitalWrite(dirRight, HIGH);
  analogWrite(mRight, valRight);
  digitalWrite(dirLeft, HIGH);
  analogWrite(mLeft, valLeft);
}

void stright(int valLeft, int valRight){
  digitalWrite(dirRight, HIGH);
  analogWrite(mRight, valLeft);
  digitalWrite(dirLeft, HIGH);
  analogWrite(mLeft, valRight);
  while(sensorVal[0]==1){readSens();}
  while(sensorVal[5]==1){readSens();}
  while(sensorVal[2]==1){readSens(); robotRun();}
}

void turnRight(int valLeft, int valRight){ 
  digitalWrite(dirRight, HIGH);
  analogWrite(mRight, valRight);
  digitalWrite(dirLeft, LOW);
  analogWrite(mLeft, valLeft);
  while(sensorVal[5]==1){readSens();}
  while(sensorVal[4]==1){readSens();} 
  while(sensorVal[3]==1){readSens();}
  while(sensorVal[2]==1){readSens();}
}

void turnLeft(int valLeft, int valRight){
  digitalWrite(dirRight, LOW);
  analogWrite(mRight, valLeft);
  digitalWrite(dirLeft, HIGH);
  analogWrite(mLeft, valRight);
  while(sensorVal[0]==1){readSens();}
  while(sensorVal[1]==1){readSens();}
  while(sensorVal[2]==1){readSens();}
  while(sensorVal[3]==1){readSens();}
}

void turnBack(int valLeft, int valRight){ 
  digitalWrite(dirRight, LOW);
  analogWrite(mRight, valRight);
  digitalWrite(dirLeft, HIGH);
  analogWrite(mLeft, valLeft); 
  while(sensorVal[2]==1 && sensorVal[3]==1){readSens();}
  while(sensorVal[2]==0 && sensorVal[3]==0){readSens();}
  while(sensorVal[2]==1 && sensorVal[3]==1){readSens();}
  while(sensorVal[4]==0){readSens();}
}

void stopRun(){
  digitalWrite(dirRight, LOW);
  analogWrite(mRight, 0);
  digitalWrite(dirLeft, LOW);
  analogWrite(mLeft, 0);  
}
